#include "common.h"
#include "dbinterface.h"

#define TIMEOUT 10

extern DbInterface dbi;
extern char *CONV_HOST;

#define DEFAULT_MIN_NODES 1
#define DEFAULT_MAX_NODES 1
#define DEFAULT_JOB_TYPE NCHARM
#define DEFAULT_ALLOCATED_TIME 14400    //14400 seconds (4hrs).

Job::Job(int nodes, int _ppn){
    status = QUEUED;
    bitmap_changed = 0;

    ip = port = pid = type = 0; //priority = 0;
    client_soc_fd = dbid = 0;

    kill_node_jobs = 0;

    start_time = 0;

    flags = 0;

    argv = NULL;
    strcpy(argbuf, "");
    strcpy(userName, "");
    strcpy(name, "");
    strcpy(script_name, "");
    strcpy(nodes_file, "");
    strcpy(email, "");
    strcpy(display, "");

    strcpy(Stdout, "socket");
    strcpy(Stderr, "socket");
    strcpy(Stdin, "socket");
    
    min_nodes = DEFAULT_MIN_NODES;
    max_nodes = DEFAULT_MAX_NODES;

    min_pe = max_pe = 0;
    type = DEFAULT_JOB_TYPE;

    allocated_time = DEFAULT_ALLOCATED_TIME;

    snprintf(working_directory, SMALL_BUF_SIZE, "%s", getenv("PWD"));

    num_system_nodes = nodes;
    num_allocated_nodes = 0;
    ppn = max_ppn = _ppn;

    memset(bit_map, 0, num_system_nodes);
    next = NULL;
}

Job::~Job(){
    
    if(!argv)
        return;
    
    //    printf("In destrcutor, before delete argv[1..\n");
    if(type == MCHARM){
        //        delete[] argv[1];
        delete[] argv[4];
    }/*
       else if(type == NCHARM)
       delete[] argv[1];
       else if(type == MPI)
       delete[] argv[2];
     */
    //    printf("In destrcutor, before delete argv\n");

    delete[] argv;
    //    printf("After destructor\n");
}

int Job::delete_all(char *system_bit_map){

    if(type == MCHARM){
        try{
            CcsFinalize(&svr);
        }
        catch(CcsException ex){
            char buf[SMALL_BUF_SIZE];
            ex.getMessage(buf);
            //            printf("Caught Exception in delete_all %s\n", buf); 
        }
    }

    for(int i=0; i < num_system_nodes; i++)
	system_bit_map[i] |= bit_map[i];

    int nfreed = num_allocated_nodes;
    num_allocated_nodes = 0;

    status = FINISHED;
    dbi.update_status(dbid, FINISHED);

    int my_userid = getuid();
    setuid(0);

    char filepath[2 * SMALL_BUF_SIZE];
    sprintf(filepath, "%s/%s", working_directory, nodes_file);
    unlink(filepath);

    if(script_name[0] != '/')
	sprintf(filepath, "%s/%s", working_directory, script_name);
    else
	sprintf(filepath, "%s", script_name);
    unlink(filepath);
    seteuid(my_userid);

    char message[LARGE_BUF_SIZE];
    time_t time_val = time(NULL);
    char timebuf[LARGE_BUF_SIZE];
    char *time_str = ctime_r(&time_val, timebuf);
    sprintf(message, "Job #%d finished at %s\n", dbid, time_str);
    
    send_email(message, NOTIFY_END);

    return nfreed;
}

int Job::connect(){
    
    //    printf("In Connect %d %d\n", dbid, port);
    
    if(status >= FAILED_CONNECT){
	printf("Job removed or already connected\n");
	return -1;
    }

    skt_ip_t sktIP = {0};
    //    sktIP.tag = ip;

    try{
        CcsConnectIpWithTimeout(&svr, sktIP, port, NULL, TIMEOUT);
    }
    catch(CcsException ex){
        printf("Caught Exception in connect\n"); 
        status = FAILED_CONNECT;
        return -1;
    }

    //Successfully connected, can do set bitmaps.
    status = CONNECTED;
    //    dbi.update_status(dbid, CONNECTED);
    return 0;
}

// Checks if the job is still alive. returns 1 if the job is alive 
// else it returns 0.
int Job::ping(int pid){
    int user_id = getuid();
    setuid(0);
    if(kill(pid, 0) != -1)
	return 1;
    else
	return 0;
    seteuid(user_id);
}

int Job::set_bitmap(){

    //    printf("In setbit map\n");

    if(type != MCHARM)
	return -1;

    // BitMap has not changed.
    if(!bitmap_changed)
	return -1;

    // Not connected or could not connect, return.
    if(status != CONNECTED) //|| (status != BITMAPSET))
      return -1;

    //    printf("set bit map alloc = %d\n", num_allocated_nodes);
    bitmap_changed = 0;
    
    //    status = BITMAPSET;

    dbi.update_bitmap(dbid, bit_map);

    int err = -1;
    try{
        char proc_bit_map[LARGE_BUF_SIZE];
        bzero(proc_bit_map, num_system_nodes * max_ppn);

        for(int ncount = 0; ncount < num_system_nodes; ncount++)
            for(int pcount = 0; pcount < ppn; pcount++)
                proc_bit_map[ncount * max_ppn + pcount] = 1;

        err = CcsSendRequestWithTimeout(&svr, "set_bitmap",  0, 
                                        num_system_nodes * max_ppn, proc_bit_map, 
                                        TIMEOUT);
    }
    catch(CcsException ex){
        printf("Caught Exception in set_bitmap\n");
    }
    return err;
}

int Job::parse_query(char *query){
    int i = 0;
    int count = 0;
    
    query[strlen(query)-1] = '\0';
   
    char tmplastsbuf[LARGE_BUF_SIZE]; 
    char *lasts = tmplastsbuf;
    count = atoi(strtok_r(query, " \n", &lasts));
    char **temp_argv = new char*[count];
    for(i = 0; i < count; i++)
	temp_argv[i] = strtok_r(NULL, " ", &lasts);
    
    /*
    type = NCHARM;
    min_nodes = 1;
    max_nodes = num_system_nodes;
    priority = 0;
    */

    int name_assigned = 0;
    int argc = 0;
    
    char tempbuf[LARGE_BUF_SIZE];
    strcpy(tempbuf, "");
    for(i = 0; i < count; i++){
	if(strncmp(temp_argv[i], "-minpe", 7) == 0){
	    if(i == count - 1){
                delete[] temp_argv;
		return -1;
            }
	    min_pe = atoi(temp_argv[i + 1]);
	    i++;
	}
	else if(strncmp(temp_argv[i], "-maxpe", 7) == 0){
	    if(i == count - 1){
                delete[] temp_argv;
		return -1;
            }
	    max_pe = atoi(temp_argv[i + 1]);
	    i++;
	}
	else if(strncmp(temp_argv[i], "-minnodes", 10) == 0){
	    if(i == count - 1){
                delete[] temp_argv;
		return -1;
            }
	    min_nodes = atoi(temp_argv[i + 1]);
	    i++;
	}
	else if(strncmp(temp_argv[i], "-maxnodes", 10) == 0){
	    if(i == count - 1){
                delete[] temp_argv;
		return -1;
            }
	    max_nodes = atoi(temp_argv[i + 1]);
	    i++;
	}
        else if(strncmp(temp_argv[i], "-ppn", 5) == 0){
	    if(i == count - 1){
                delete[] temp_argv;
		return -1;
            }
	    ppn = atoi(temp_argv[i + 1]);
            printf("Found ppn = %d\n", ppn);
	    i++;
	}
	else if(strncmp(temp_argv[i], "-type", 6) == 0){
	    if(i == count - 1){
                delete[] temp_argv;
		return -1;
            }
	    if(strncasecmp(temp_argv[i + 1], "mcharm", 7) == 0)
		type = MCHARM;
	    else if(strncasecmp(temp_argv[i + 1], "mpi", 4) == 0)
		type = MPI;
	    else if(strncasecmp(temp_argv[i + 1], "charm", 6) == 0)
		type = NCHARM;
	    else if(strncasecmp(temp_argv[i + 1], "uni", 4) == 0)
		type = UNI;
	    else {
                delete[] temp_argv;
                return -1;
            }
	    i++;
	}
	/*
	else if(strncmp(temp_argv[i], "-priority", 10) == 0){
	    if(i == count - 1)
		return -1;
	    priority = atoi(temp_argv[i + 1]);
	    i++;
	}
	*/
	else if(strncmp(temp_argv[i], "-stdout", 7) == 0){
	    if(i == count - 1){
                delete[] temp_argv;
		return -1;
            }
	    snprintf(Stdout, SMALL_BUF_SIZE, "%s", temp_argv[i + 1]);
	    i++;
	}
	else if(strncmp(temp_argv[i], "-stdin", 6) == 0){
	    if(i == count - 1){
                delete[] temp_argv;
		return -1;
            }
	    snprintf(Stdin, SMALL_BUF_SIZE, "%s", temp_argv[i + 1]);
	    i++;
	}
	else if(strncmp(temp_argv[i], "-stderr", 7) == 0){
	    if(i == count - 1){
                delete[] temp_argv;
		return -1;
            }
	    snprintf(Stderr, SMALL_BUF_SIZE, "%s", temp_argv[i + 1]);
	    i++;
	}
	else if(strncmp(temp_argv[i], "-pwd", 5) == 0){
	    if(i == count - 1){
                delete[] temp_argv;
		return -1;
            }
	    snprintf(working_directory, SMALL_BUF_SIZE, "%s", temp_argv[i + 1]);
	    i++;
	}
	else if(strncmp(temp_argv[i], "-time", 6) == 0){
	    if(i == count - 1){
                delete[] temp_argv;
		return -1;
            }

	    int val1 = 0, val2 = 0, val3 = 0;
            
	    int valcount = sscanf(temp_argv[i + 1], "%d:%d:%d", &val1, &val2, &val3);

	    if(valcount == 1)
		allocated_time = val1;
	    else if(valcount == 2)
		allocated_time = val1 * 60 + val2;
	    else if(valcount == 3)
		allocated_time = val1 * 3600 + val2 * 60 + val3;
	    
	    //printf("Allocated Time %d, valcount = %d\n", allocated_time, valcount); 

	    i++;
	}
        else if(strncmp(temp_argv[i], "-email", 7) == 0){
            
            if(i == count - 1){
                delete[] temp_argv;
		return -1;
            }
            
            snprintf(email, SMALL_BUF_SIZE, "%s", temp_argv[i + 1]);
            i++;
            flags |= NOTIFY_END;
        }
        else if(strncmp(temp_argv[i], "-display", 9) == 0){
            
            if(i == count - 1){
                delete[] temp_argv;
		return -1;
            }
            
            snprintf(display, SMALL_BUF_SIZE, "%s", temp_argv[i + 1]);
            i++;
        }
        else if(strncmp(temp_argv[i], "++debug", 8) == 0){
            flags |= DEBUG_ON;
        }
        else if(strncmp(temp_argv[i], "++notify-start", 15) == 0){
            flags |= NOTIFY_START;
        }
        else if(strncmp(temp_argv[i], "++notify-end", 14) == 0){
            flags |= NOTIFY_END;
        }
        else if(strncmp(temp_argv[i], "++server", 9) == 0){
            flags |= CCS_ON;
        }
	else if(!name_assigned){
	    strncpy(name, temp_argv[i], SMALL_BUF_SIZE);
	    name[SMALL_BUF_SIZE-1] = '\0';
	    name_assigned = 1;
	}
	else{
	    strncat(tempbuf, temp_argv[i], LARGE_BUF_SIZE);
	    strcat(tempbuf, " ");                 // New seperator later.
	    argc ++;
	}
    }
    
    if(!name_assigned){
        delete[] temp_argv;
	return -1;
    }

    if(ppn <= 0)
        ppn = max_ppn;

    if(max_pe < min_pe)
	max_pe = min_pe;

    if(max_pe < ppn && max_pe != 0)
	ppn = max_pe;

    if(min_pe / ppn > min_nodes)
        min_nodes = min_pe / ppn;

    if(max_pe / ppn > max_nodes)
        max_nodes = max_pe / ppn;

    if(type == UNI)
	min_nodes = max_nodes = 1;
    
    if(min_nodes < 1)
        min_nodes = 1;

    if(max_nodes < 1)
        max_nodes = 1;

    if(min_nodes > max_nodes)
	max_nodes = min_nodes;
    
    min_pe = min_nodes * ppn;
    max_pe = max_nodes * ppn;

    snprintf(argbuf, LARGE_BUF_SIZE, "%s", tempbuf);
    delete[] temp_argv;

    //    printf("After parse query\n");

    return 0;
}

void Job::re_init(){
    sprintf(nodes_file, ".nodelist.%d", dbid);
    if(type == UNI)
        sprintf(script_name, "%s/.script.%d", working_directory, dbid);
    else
        sprintf(script_name, ".script.%d", dbid);
}

void Job::init_arg(){    
    
    printf("In Init Arg\n");
    sprintf(nodes_file, ".nodelist.%d", dbid);
    if(type == UNI)
        sprintf(script_name, "%s/.script.%d", working_directory, dbid);
    else
        sprintf(script_name, ".script.%d", dbid);
    
    int argcount = 0;	
    // We prepare the job's command-line arguments here.
    int nargs = 0;
    if(type == MCHARM){
        
        nargs = 8 + 1;
        /*
        if(flags & DEBUG_ON)
            nargs ++;
        */
	argv = new char*[nargs];
	argv[0] = CONV_HOST;
	snprintf(num_proc_string, VERY_SMALL_BUF_SIZE, "+p%d", 
                 num_system_nodes * max_ppn);
        
	argv[1] = num_proc_string;
	argv[2] = "++server";
	argv[3] = "++server-port";
	argv[4] = new char[VERY_SMALL_BUF_SIZE];
	snprintf(argv[4], VERY_SMALL_BUF_SIZE, "%d", port);
	argv[5] = "++nodelist";
	argv[6] = nodes_file;

	argcount += 7;
        /*
        if(flags & DEBUG_ON)
            argv[argcount ++] = "++debug";
        */
    }
    else if(type == MPI){
        
        nargs = 7 + 1;

	argv = new char*[nargs];
	argv[0] = MPIRUN;
	argv[1] = "-np";	
	snprintf(num_proc_string, VERY_SMALL_BUF_SIZE, "%d", 
                 num_allocated_nodes * ppn);
	argv[2] = num_proc_string;
	argv[3] = "-machinefile";
	argv[4] = nodes_file;
	argv[5] = "-nolocal";
	argcount += 6;
        
        kill_node_jobs = 1;
    }
    else if(type == NCHARM){
        kill_node_jobs = 1;
        nargs = 5 + 1;
        
        /*
          if(flags & DEBUG_ON)
          nargs ++;
        */

        if(flags & CCS_ON)
            nargs ++;

	argv = new char*[nargs];
	argv[0] = CONV_HOST;
        printf("PPN for the job = %d\n", ppn);
	snprintf(num_proc_string, VERY_SMALL_BUF_SIZE, "+p%d", 
                 num_allocated_nodes * ppn);
	argv[1] = num_proc_string;
	argv[2] = "++nodelist";
	argv[3] = nodes_file;
	argcount += 4;
        
        /*
        if(flags & DEBUG_ON)
            argv[argcount ++] = "++debug";
        */

        if(flags & CCS_ON)
            argv[argcount ++] = "++server";
    }
    else if(type == UNI){
        kill_node_jobs = 1;
        printf("job type is UNI\n");

	argv = new char*[3 + 1];
	argv[0] = RSH;
        //	delete[] num_proc_string;
	argv[1] = NULL;
	argcount += 2;
    }
    argv[argcount++] = script_name; //script name;

    //if(argcount > 2)
    //  printf("init_arg %s, %s, %s\n", argv[0], argv[1], argv[2]);
    
    argv[argcount] = NULL;
    printf("Out of init arg\n");
}

void Job::add_node(int n){
    
    num_allocated_nodes++;
    if(n < num_system_nodes)
	bit_map[n] = 1;
    bitmap_changed = 1;
}

int Job::delete_node(){
    num_allocated_nodes --;
    int i;
    bitmap_changed = 1;
    for(i=0; i < num_system_nodes; i++)
	if(bit_map[i] == 1)
	    break;

    if(bit_map[i] == 1){
	bit_map[i] = 0;
	return i;
    }
    else return -1;
}

void Job::Kill(){
    int user_id = getuid();
    setuid(0);
    printf("Killing %d\n", pid);
    kill(pid, 9);
    seteuid(user_id);
}

void Job::started(){

    status = RUNNING;
    dbi.update_pid(dbid, pid);
    dbi.update_bitmap(dbid, bit_map);
    dbi.update_status(dbid, RUNNING);
    start_time = (int)getCurrentTime();
    dbi.update_startTime(dbid, start_time);
}

void Job::send_email(const char* message, int state){
    if((flags & NOTIFY_START & state) || (flags & NOTIFY_END & state)){
        char address[SMALL_BUF_SIZE];
        if(strcmp(email, "") == 0)
            sprintf(address, "%s@localhost", userName);
        else
            strcpy(address, email);

        char command[LARGE_BUF_SIZE];
        char buf[SMALL_BUF_SIZE];

        sprintf(buf, "/bin/touch .sendmail.%d;", dbid);
        strcpy(command, buf);
        sprintf(buf, "/bin/echo \"%s\" >> .sendmail.%d;", message, dbid);
        strcat(command, buf);
        sprintf(buf, "/bin/mail -s \"Faucets Job Report\" %s < .sendmail.%d;", address, dbid);
        strcat(command, buf);
        sprintf(buf, "/bin/rm -f .sendmail.%d;", dbid);
        strcat(command, buf);
        
        system(command);
    }
}

void* Job::operator new (size_t s)
{
    //  std::cout << " [ new      Bytes  = " << s << " ]" << "\n";
    return malloc(s);            // The global new 
}

void Job::operator delete(void *p, size_t s)
{
    //  std::cout << " [ delete   Bytes = " << s << " ]" << "\n";
    free(p);                               // The global delete 
}

